# Copyright 2004 James Bunton <james@delx.cjb.net>
# Licensed for distribution under the GPL version 2, check COPYING for details

from twisted.internet import reactor
from twisted.internet.protocol import ClientFactory
from tlib.domish import parseText
from tlib import msn
import config
import utils
import debug




class MSNConnection:
	""" Manages all the Twisted factories, etc """
	def __init__(self, username, password):
		self.username = username
		self.password = password
		self.inited = False
		self.tries = 0
		self.initMe()
		debug.log("MSNConnection: \"%s\" created" % (self.username))
		
	def initMe(self):
		if(self.inited):
			MSNConnection.removeMe(self)
		
		self.switchboardSessions = {}
		dispatchFactory = DispatchFactory(self)
		reactor.connectTCP('messenger.hotmail.com', 1863, dispatchFactory)
		self.notificationFactory = msn.NotificationFactory()
		self.notificationFactory.userHandle = self.username
		self.notificationFactory.password = self.password
		self.notificationFactory.msncon = self
		self.notificationFactory.protocol = Notification
		self.notificationFactory.initialListVersion = self.initialListVersion
		self.notificationProtocol = None
		
		self.savedStatus = None
		
		self.inited = True
		
		debug.log("MSNConnection: \"%s\" initialised" % (self.username))
	
	def removeMe(self):
		debug.log("MSNConnection: \"%s\" destroyed" % (self.username))
		if(self.notificationProtocol):
			self.notificationProtocol.removeMe()
		self.notificationFactory = None
		self.notificationProtocol = None
		for userHandle in utils.copyDict(self.switchboardSessions):
			self.switchboardSessions[userHandle].removeMe()
		self.switchboardSessions = {}
	
	def resourceOffline(self, offlineResource):
		for contact in self.switchboardSessions.keys():
			if(self.switchboardSessions[contact].resource == offlineResource):
				self.switchboardSessions[contact].resource = self.highestResource()
	
	def getContacts(self):
		if(self.notificationFactory):
			return self.notificationFactory.contacts
		else:
			return None
		
	
	def sendMessage(self, remoteUser, resource, text):
		debug.log("MSNConnection: \"%s\" sendMessage(\"%s\", \"%s\")" % (self.username, remoteUser, text))
		if(not self.switchboardSessions.has_key(remoteUser)):
			self.switchboardSessions[remoteUser] = SwitchboardSession(self, remoteUser, resource)
		self.switchboardSessions[remoteUser].resource = resource
		self.switchboardSessions[remoteUser].sendMessage(text)
	
	def sendTypingToContact(self, remoteUser):
		if(self.switchboardSessions.has_key(remoteUser)):
			self.switchboardSessions[remoteUser].sendTypingNofication()
	
	def notificationProtocolReady(self, notificationProtocol):
		self.notificationProtocol = notificationProtocol
		self.loggedIn()
		self.tries = 0
	
	def sendSavedStatus(self):
		# Hack for initial status
		if(self.savedStatus):
			statusCode, screenName = self.savedStatus
			self.savedStatus = None
			self.changeStatus(statusCode, screenName)
	
	def changeStatus(self, statusCode, screenName):
		if(self.notificationProtocol):
			def cb1(arg):
				self.ourStatusChanged(arg[0])
			def cb2(arg):
				self.ourNickChanged(arg[0])
			debug.log("MSNConnection: \"%s\" - changing status and screenName (\"%s\", \"%s\")" % (self.username, statusCode, screenName))
			statusCode = str(statusCode.encode("utf-8"))
			screenName = str(screenName.encode("utf-8"))
			self.notificationProtocol.changeStatus(statusCode).addCallback(cb1)
			self.notificationProtocol.changeScreenName(screenName).addCallback(cb2)
		else:
			self.savedStatus = (statusCode, screenName)
	
	def connectionLostBase(self, reason):
		# Attempts to reconnect
		if(self.tries < 5):
			reactor.callLater(2 ** self.tries, self.initMe)
			self.tries += 1
		else:
			self.removeMe()
			self.connectionLost(self)
	
	def addContact(self, listType, userHandle):
		return self.notificationProtocol.addContact(listType, str(userHandle))
	
	def remContact(self, listType, userHandle, groupID=0):
		return self.notificationProtocol.remContact(listType, str(userHandle))
	
	
	
	def initialEmailNotification(self, inboxunread, foldersunread):
		pass
	
	def realtimeEmailNotification(self, mailfrom, fromaddr, subject):
		pass
	
	def loggedIn(self):
		pass
	
	def loginFailure(self, message):
		pass
		
	def multipleLogin(self):
		pass
	
	def gotMessage(self, remoteUser, resource, text):
		pass
	
	def listSynchronized(self):
		pass
	
	def contactStatusChanged(self, remoteUser):
		pass
	
	def ourStatusChanged(self, statusCode):
		pass
	
	def userMapping(self, passport, jid):
		pass
	
	def gotContactTyping(self, remoteUser, resource):
		pass
	
	def ourNickChanged(self, arg):
		pass
	
	def serverGoingDown(self):
		pass
	
	def accountNotVerified(self):
		pass
	
	def userAddedMe(self, userHandle):
		pass
	
	def userRemovedMe(self, userHandle):
		pass
	
	def failedMessage(self, remoteUser):
		pass
	
	def connectionLost(self):
		pass




def switchToGroupchat(switchboardSession, user1, user2):
	gcsbs = GroupchatSwitchboardSession()
	from glue import LegacyGroupchat, msn2jid
	groupchat = LegacyGroupchat(session=switchboardSession.msncon.session, resource=None, existing=True, switchboardSession=gcsbs)
	gcsbs.groupchat = groupchat
	gcsbs.msncon = switchboardSession.msncon
	gcsbs.switchboard = switchboardSession.switchboard
	gcsbs.switchboard.switchboardSession = gcsbs
	gcsbs.ready = True
	gcsbs.userJoined(user1)
	gcsbs.userJoined(user2)
	groupchat.sendUserInvite(msn2jid(switchboardSession.remoteUser))
	switchboardSession.removeMe(False)
	debug.log("GroupchatSwitchboardSession: \"%s\" \"%s\" created by conversion" % (gcsbs.groupchat.roomJID(), gcsbs))
	return gcsbs


class GroupchatSwitchboardSession:
	def __init__(self, groupchat=None, makeSwitchboard=False):
		self.msncon = None
		self.groupchat = None
		if(groupchat):
			self.groupchat = groupchat
			self.msncon = self.groupchat.session.legacycon
		self.switchboard = None
		self.ready = False
		self.messageBuffer = []
		self.invitedUsers = []
		
		if(makeSwitchboard and groupchat):
			debug.log("GroupchatSwitchboardSession: \"%s\" \"%s\" requesting a switchboard session" % (self.groupchat.roomJID(), self))
			self.msncon.notificationProtocol.requestSwitchboardServer().addCallback(self.sbRequestAccepted)
		
		if(self.msncon):
			debug.log("GroupchatSwitchboardSession: \"%s\" \"%s\" created" % (self.msncon.username, self))
	
	def removeMe(self):
		debug.log("GroupchatSwitchboardSession: \"%s\" \"%s\" destroyed" % (self.groupchat.roomJID(), self))
		self.msncon = None
		self.switchboard.removeMe()
		self.switchboard = None
		self.groupchat = None
		self.ready = False

		utils.mutilateMe(self)
	
	def sbRequestAccepted(self, (host, port, key)):
		# Connect to the switchboard server
		debug.log("GroupchatSwitchboardSession: \"%s\" \"%s\" sbRequestAccepted()" % (self.msncon.username, self))
		reactor.connectTCP(host, port, SwitchboardFactory(self, key))
	
	def sendMessage(self, message):
		if(self.ready):
			message = str(message.encode("utf-8"))
			msnmessage = msn.MSNMessage(message=message)
			msnmessage.setHeader("Content-Type", "text/plain; charset=UTF-8")
			msnmessage.ack = msn.MSNMessage.MESSAGE_ACK_NONE
			self.switchboard.sendMessage(msnmessage)
		else:
			self.messageBuffer.append(message)
	
	def inviteUser(self, userHandle):
		userHandle = str(userHandle)
		if(self.ready):
			debug.log("GroupchatSwitchboardSession: \"%s\" \"%s\" inviting %s" % (self.msncon.username, self, userHandle))
			self.switchboard.inviteUser(userHandle)
		else:
			self.invitedUsers.append(userHandle)
	
	def gotMessage(self, message):
		self.groupchat.messageReceived(message.userHandle, message.getMessage())
	
	def flushBuffer(self):
		for m in utils.copyList(self.messageBuffer):
			self.messageBuffer.remove(m)
			self.sendMessage(m)
		
		for i in self.invitedUsers:
			self.inviteUser(i)
	
	def userJoined(self, userHandle):
		debug.log("GroupchatSwitchboardSession: \"%s\" \"%s\" userJoined(\"%s\")" % (self.msncon.username, self, userHandle))
		self.groupchat.contactJoined(userHandle)
	
	def userLeft(self, userHandle):
		debug.log("GroupchatSwitchboardSession: \"%s\" \"%s\" userLeft(\"%s\")" % (self.msncon.username, self, userHandle))
		self.groupchat.contactLeft(userHandle)



class SwitchboardSession:
	def __init__(self, msncon, remoteUser, resource, reply=False, host=None, port=None, key=None, sessionID=None):
		self.msncon = msncon
		self.remoteUser = str(remoteUser)
		self.resource = str(resource)
		
		self.killTimer = reactor.callLater(30.0*60.0, self.removeMe)
		
		self.switchboard = None # The SwitchboardClient class
		self.messageBuffer = [] # Any messages sent before the switchboard is ready are buffered
		self.ready = False # Is True when we are connected to the switchboard, and the remote user has accepted our invite
		
		if(not reply):
			# Request a switchboard
			self.msncon.notificationProtocol.requestSwitchboardServer().addCallback(self.sbRequestAccepted)
		else:
			reactor.connectTCP(host, port, SwitchboardFactory(self, key, sessionID, reply))
		
		debug.log("SwitchboardSession: \"%s\" \"%s\" \"%s\" created" % (self.msncon.username, self.remoteUser, self.resource))
	
	def removeMe(self, sbflag=True):
		debug.log("SwitchboardSession: \"%s\" \"%s\" \"%s\" destroyed" % (self.msncon.username, self.remoteUser, self.resource))
		del self.msncon.switchboardSessions[self.remoteUser]
		self.msncon = None
		if(sbflag and self.switchboard):
			self.switchboard.removeMe()
		self.switchboard = None
		self.ready = False
		if(self.killTimer and not self.killTimer.called):
			self.killTimer.cancel()
		self.killTimer = None

		utils.mutilateMe(self)
	
	def resetTimer(self):
		# Sets a count down timer to kill this switchboard session in 30 minutes
		self.killTimer.cancel()
		self.killTimer = reactor.callLater(30.0*60.0, self.removeMe)
	
	def tearDownFail(self):
		# Notify the user that any buffered messages have failed, and tear down the switchboard session
		debug.log("SwitchboardSession: \"%s\" \"%s\" tearing down because remote user ignored our invitation" % (self.msncon.username, self.remoteUser))
		for message in self.messageBuffer:
			self.msncon.failedMessage(self.remoteUser, message)
		self.removeMe()
	
	def sbRequestAccepted(self, (host, port, key)):
		# Connect to the switchboard server
		reactor.connectTCP(host, port, SwitchboardFactory(self, key))
	
	def sendMessage(self, message):
		if(self.ready):
			debug.log("SwitchboardSession: \"%s\" \"%s\" sending message \"%s\"" % (self.msncon.username, self.remoteUser, message))
			message = str(message.encode("utf-8"))
			msnmessage = msn.MSNMessage(message=message)
			msnmessage.setHeader("Content-Type", "text/plain; charset=UTF-8")
			msnmessage.ack = msn.MSNMessage.MESSAGE_NACK
			def failedMessage(ignored):
				self.msncon.failedMessage(self.remoteUser, message)
			d = self.switchboard.sendMessage(msnmessage)
			d.addErrback(failedMessage)
			self.resetTimer()
		else:
			self.messageBuffer.append(message)
	
	def sendTypingNofication(self):
		if(self.ready):
			self.switchboard.sendTypingNotification()
	
	def contactTyping(self):
		self.msncon.gotContactTyping(self.remoteUser, self.resource)
	
	def flushBuffer(self):
		for m in utils.copyList(self.messageBuffer):
			self.messageBuffer.remove(m)
			self.sendMessage(m)
	
	def gotMessage(self, message):
		self.msncon.gotMessage(self.remoteUser, self.resource, message.getMessage())
		self.resetTimer()
	
	def userJoined(self, userHandle):
		if(userHandle != self.remoteUser):
			# Another user has joined, so we now have three participants (these two and ourself)
			switchToGroupchat(self, self.remoteUser, userHandle)
	
	def userLeft(self, userHandle):
		if(userHandle == self.remoteUser):
			self.removeMe()






class DispatchFactory(ClientFactory):
	def __init__(self, msncon):
		self.msncon = msncon
	
	def buildProtocol(self, addr):
		p = Dispatch(self.msncon)
		del self.msncon # No longer needed
		return p


class Dispatch(msn.DispatchClient):
	def __init__(self, msncon):
		msn.DispatchClient.__init__(self)
		self.msncon = msncon
		self.userHandle = self.msncon.username
	
	def __del__(self):
		self.factory = None
		self.msncon = None
	
	def gotNotificationReferral(self, host, port):
		self.transport.loseConnection()
		if(self.msncon and self.msncon.session and self.msncon.session.alive):
			reactor.connectTCP(host, port, self.msncon.notificationFactory)



class Notification(msn.NotificationClient):
	def __init__(self):
		msn.NotificationClient.__init__(self, proxy=config.proxyServer, proxyport=config.proxyPort)

	def removeMe(self):
		self.logOut()
		self.transport.loseConnection()
		self.factory.msncon.notificationProtocol = None
		self.factory.msncon = None
		self.factory = None

		utils.mutilateMe(self)
	
	def loginFailure(self, message):
		debug.log("NotificationClient: \"%s\" created" % (self.factory.msncon.username))
		self.factory.msncon.loginFailure(message)
	
	def loggedIn(self, userHandle, screenName, verified):
		self.factory.msncon.notificationProtocolReady(self)
		if(not verified):
			self.factory.msncon.accountNotVerified()
		
		msn.NotificationClient.loggedIn(self, userHandle, screenName, verified)
		
		debug.log("NotificationClient: \"%s\" authenticated with MSN servers" % (self.factory.msncon.username))
	
	def gotMessage(self, msnmessage):
		debug.log("NotificationClient: \"%s\" gotMessage()" % (self.factory.msncon.username))
	
		cTypes = [s.lstrip() for s in msnmessage.getHeader("Content-Type").split(';')]
		def getFields():
			fields = msnmessage.getMessage().strip().split('\n')
			values = {}
			for i in fields:
				a = i.split(':')
				if(len(a) != 2): continue
				f, v = a
				f = f.strip()
				v = v.strip()
				values[f] = v
			return values
		
		if("text/x-msmsgsinitialemailnotification" in cTypes and config.mailNotifications):
			values = getFields()
			try:
				inboxunread = int(values["Inbox-Unread"])
				foldersunread = int(values["Folders-Unread"])
			except KeyError:
				return
			if(foldersunread + inboxunread == 0): return # For some reason MSN sends notifications about empty inboxes sometimes?
			debug.log("NotificationClient: \"%s\" Initial hotmail notification" % (self.factory.msncon.username))
			self.factory.msncon.initialEmailNotification(inboxunread, foldersunread)
		
		elif("text/x-msmsgsemailnotification" in cTypes and config.mailNotifications):
			values = getFields()
			try:
				mailfrom = values["From"]
				fromaddr = values["From-Addr"]
				subject = values["Subject"]
				junkbeginning = "=?\"us-ascii\"?Q?"
				junkend = "?="
				subject = subject.replace(junkbeginning, "").replace(junkend, "").replace("_", " ")
			except KeyError:
				# If any of the fields weren't found then it's not a big problem. We just ignore the message
				return
			debug.log("NotificationClient: \"%s\" Live hotmail notification" % (self.factory.msncon.username))
			self.factory.msncon.realtimeEmailNotification(mailfrom, fromaddr, subject)

		elif("NOTIFICATION" == msnmessage.userHandle):
			notification = parseText(msnmessage.message)
			siteurl = notification.getAttribute("siteurl")
			notid = notification.getAttribute("id")
			msg = notification.getElement("MSG")
			msgid = item.getAttribute("id")
			action = msg.getElement("ACTION").getAttribute("url")
			subscr = msg.getElement("SUBSCR").getAttribute("url")
			bodytext = msg.getElement("BODY").getElement("TEXT")
			
			actionurl = "%s//%s&notification_id=%s&message_id=%s&agent=messenger" % (siteurl, action, notid, msgid)
			subscrurl = "%s//%s&notification_id=%s&message_id=%s&agent=messenger" % (siteurl, subscr, notid, msgid)
			
			self.factory.msncon.msnAlert(bodytext, actionurl, subscrurl)

	
	
	def connectionLost(self, reason):
		if(self.factory == None): return
		debug.log("NotificationClient: \"%s\" lost connection with MSN servers" % (self.factory.userHandle))
		msn.NotificationClient.connectionLost(self, reason)
		self.factory.msncon.connectionLostBase(reason)
		self.removeMe()
	
	def listSynchronized(self, *args):
		debug.log("NotificationClient: \"%s\" MSN contact lists synchronised" % (self.factory.userHandle))
		self.factory.msncon.listSynchronized()
		self.factory.msncon.sendSavedStatus()
	
	def gotSwitchboardInvitation(self, sessionID, host, port, key, remoteUser, screenName):
		debug.log("NotificationClient: \"%s\" gotSwitchboardInvitation(\"%s\")" % (self.factory.userHandle, remoteUser))
		sbs = SwitchboardSession(self.factory.msncon, remoteUser, self.factory.msncon.session.highestResource(), True, host, port, key, sessionID)
		self.factory.msncon.switchboardSessions[remoteUser] = sbs
	
	def contactStatusChanged(self, statusCode, userHandle, screenName):
		debug.log("NotificationClient: \"%s\" contactStatusChanged(\"%s\", \"%s\")" % (self.factory.userHandle, statusCode, userHandle))
		msn.NotificationClient.contactStatusChanged(self, statusCode, userHandle, screenName)
		
		self.factory.msncon.contactStatusChanged(userHandle)
	
	def gotContactStatus(self, statusCode, userHandle, screenName):
		msn.NotificationClient.gotContactStatus(self, statusCode, userHandle, screenName)
		debug.log("NotificationClient: \"%s\" gotContactStatus(\"%s\", \"%s\")" % (self.factory.userHandle, statusCode, userHandle))
		
		self.factory.msncon.contactStatusChanged(userHandle)
	
	def contactOffline(self, userHandle):
		debug.log("NotificationClient: \"%s\" contactOffline(\"%s\")" % (self.factory.userHandle, userHandle))
		msn.NotificationClient.contactOffline(self, userHandle)
		
		self.factory.msncon.contactStatusChanged(userHandle)
	
	def userAddedMe(self, userHandle, screenName, listVersion):
		debug.log("NotificationClient: \"%s\" userAddedMe(\"%s\", \"%s\")" % (self.factory.userHandle, userHandle, listVersion))
		msn.NotificationClient.userAddedMe(self, userHandle, screenName, listVersion)
		self.factory.msncon.userAddedMe(userHandle)
	
	def userRemovedMe(self, userHandle, listVersion):
		debug.log("NotificationClient: \"%s\" userRemovedMe(\"%s\", \"%s\")" % (self.factory.userHandle, userHandle, listVersion))
		msn.NotificationClient.userRemovedMe(self, userHandle, listVersion)
		self.factory.msncon.userRemovedMe(userHandle)
	
	def multipleLogin(self):
		debug.log("NotificationClient: \"%s\" multiple logins" % (self.factory.msncon.username))
		self.factory.msncon.multipleLogin()



class SwitchboardFactory(ClientFactory):
	def __init__(self, switchboardSession, key, sessionID=None, reply=False):
		self.switchboardSession = switchboardSession
		self.key = key
		self.sessionID = sessionID
		self.reply = reply
	
	def buildProtocol(self, addr):
		p = Switchboard(self.switchboardSession)
		p.key = self.key
		p.sessionID = self.sessionID
		p.reply = self.reply
		p.userHandle = self.switchboardSession.msncon.username
		p.factory = self
		return p

class Switchboard(msn.SwitchboardClient):
	def __init__(self, switchboardSession):
		msn.SwitchboardClient.__init__(self)
		self.switchboardSession = switchboardSession
		self.chattingUsers = []
		self.callid = None
		debug.log("SwitchboardClient: \"%s\" \"%s\" - created" % (self.switchboardSession.msncon.username, self.switchboardSession))
	
	def removeMe(self):
		self.transport.loseConnection()
		debug.log("SwitchboardClient: \"%s\" - destroyed" % (self.switchboardSession))
		self.switchboardSession = None
		self.factory.switchboardSession = None
		self.factory = None

		utils.mutilateMe(self)
	
	def loggedIn(self):
		if((not self.reply) and self.switchboardSession.__class__ == SwitchboardSession):
			def cb(arg=None):
				debug.log("SwitchboardClient: \"%s\" \"%s\" - user has joined" % (self.switchboardSession.msncon.username, self.switchboardSession.remoteUser))
			def failCB(arg=None):
				if(not self.switchboardSession.ready):
					debug.log("SwitchboardClient: \"%s\" \"%s\" - user has NOT joined after 30 seconds" % (self.switchboardSession.msncon.username, self.switchboardSession.remoteUser))
					self.switchboardSession.tearDownFail()
			d = self.inviteUser(self.switchboardSession.remoteUser)
			d.addCallback(self.delayReadySwitchboardSession)
			d.addCallback(cb)
			d.addErrback(failCB)
			# If the user doesn't join then we want to tear down the SwitchboardSession
			self.callid = reactor.callLater(30.0, failCB)
		
		else:
			self.readySwitchboardSession()
	
	def delayReadySwitchboardSession(self, ignored=None):
		reactor.callLater(3.0, self.readySwitchboardSession)
	
	def readySwitchboardSession(self, ignored=None):
		debug.log("SwitchboardClient: \"%s\" \"%s\" - ready for use" % (self.switchboardSession.msncon.username, self.switchboardSession))
		self.switchboardSession.ready = True
		self.switchboardSession.switchboard = self
		self.switchboardSession.flushBuffer()
		for user in self.chattingUsers:
			self.switchboardSession.userJoined(user)
		if(self.callid):
			self.callid.cancel() # Cancel the invite fail message (only applies if we needed to invite the user)
	
	def gotChattingUsers(self, users):
		for user in users:
			self.chattingUsers.append(user)
	
	def userJoined(self, userHandle, screenName):
		debug.log("SwitchboardClient: \"%s\" \"%s\" - userJoined(\"%s\")" % (self.switchboardSession.msncon.username, self.switchboardSession, userHandle))
		self.switchboardSession.userJoined(userHandle)
		self.sendClientCaps()
	
	def userLeft(self, userHandle):
		debug.log("SwitchboardClient: \"%s\" \"%s\" - userLeft(\"%s\")" % (self.switchboardSession.msncon.username, self.switchboardSession, userHandle))
		self.switchboardSession.userLeft(userHandle)
	
	def gotMessage(self, message):
		debug.log("SwitchboardClient: \"%s\" \"%s\" gotMessage(\"%s\")" % (self.switchboardSession.msncon.username, message.userHandle, message.getMessage()))
		cTypes = [s.lstrip() for s in message.getHeader("Content-Type").split(';')]
		if("text/plain" in cTypes):
			self.switchboardSession.gotMessage(message)
		elif("text/x-clientcaps" in cTypes):
			if(message.hasHeader("JabberID")):
				jid = message.getHeader("JabberID")
				self.switchboardSession.msncon.userMapping(message.userHandle, jid)
		else:
			debug.log("THROWING AWAY MESSAGE: %s" % (message.getMessage()))
	
	def userTyping(self, message):
		if(self.switchboardSession.__class__ == SwitchboardSession): # Ignore typing in groupchats
			if(message.userHandle == self.switchboardSession.remoteUser):
				self.switchboardSession.contactTyping()
	
	def sendClientCaps(self):
		message = msn.MSNMessage()
		message.setHeader("Content-Type", "text/x-clientcaps")
		message.setHeader("Client-Name", "PyMSNt")
		message.setHeader("JabberID", str(self.switchboardSession.msncon.session.jabberID)) # FIXME, this is a little deep
		self.sendMessage(message)

 
