# Copyright 2004 James Bunton <james@delx.cjb.net>
# Licensed for distribution under the GPL version 2, check COPYING for details

from twisted.internet.ssl import ClientContextFactory
from twisted.internet.protocol import Protocol, ClientFactory
from twisted.protocols.http import HTTPClient
from twisted.internet import reactor


class ProxyFactory(ClientFactory):
	def __init__(self, server, serverport, otherfac):
		self.server, self.serverport, self.otherfac = server, serverport, otherfac

	def buildProtocol(self, addr):
		p = ProxyProtocol()
		p.server, p.serverport, p.otherfac = self.server, self.serverport, self.otherfac
		return p

class ProxyProtocol(HTTPClient):
	def connectionMade(self):
		self.ctxFactory = ClientContextFactory()
		self.sendCommand("CONNECT", self.server + ":" + str(self.serverport))
		self.endHeaders()
		debug.log("proxy_connect_ssl: Sent CONNECT")
	
	def handleStatus(self, version, status, message):
		if(status == "200"): # Excellent
			# Must flip over into SSL now
			debug.log("proxy_connect_ssl: Flipping into SSL mode")
			self.transport.startTLS(self.ctxFactory)
			p = self.otherfac.buildProtocol(self.transport.getPeer())
			p.makeConnection(self.transport)
			self.transport.protocol = p
			p.connectionMade()
			debug.log("proxy_connect_ssl: Flipped into SSL mode successfully. Handed over to real protocol.")
		else:
			# Must cancel factory
			debug.log("proxy_connect_ssl: Error with proxy")
			self.otherfac.clientConnectionFailed(self.transport.connector, "Proxy connection failed")

	def handleHeader(self, key, val):
		pass
	
	def handleEndHeaders(self):
		pass
	
	def handleResponse(self, r):
		pass

def proxy_connect_ssl(proxy, proxyport, server, serverport, factory):
	debug.log("proxy_connect_ssl")
	pfac = ProxyFactory(server, serverport, factory)
	reactor.connectTCP(proxy, proxyport, pfac)


